Class: Gitlab::Triage::Engine

Inherits:
Object
  • Object
show all
Defined in:
lib/gitlab/triage/engine.rb

Constant Summary collapse

FILTER_MAP =

This filter map is used to help make the filter_resource method smaller. We loop through each of the keys (conditions) and map that to the filters that will be used for it.

{
  date: {
    'branches' => Filters::BranchDateFilter,
    'issues' => Filters::IssueDateConditionsFilter,
    'merge_requests' => Filters::MergeRequestDateConditionsFilter
  },
  protected: Filters::BranchProtectedFilter,
  assignee_member: Filters::AssigneeMemberConditionsFilter,
  author_member: Filters::AuthorMemberConditionsFilter,
  discussions: Filters::DiscussionsConditionsFilter,
  no_additional_labels: Filters::NoAdditionalLabelsConditionsFilter,
  ruby: Filters::RubyConditionsFilter,
  votes: Filters::VotesConditionsFilter,
  upvotes: Filters::VotesConditionsFilter,
  work_item_status: Filters::WorkItemStatusConditionsFilter
}.freeze
DEFAULT_NETWORK_ADAPTER =
Gitlab::Triage::NetworkAdapters::HttpartyAdapter
DEFAULT_GRAPHQL_ADAPTER =
Gitlab::Triage::NetworkAdapters::GraphqlAdapter
ALLOWED_STATE_VALUES =
{
  issues: %w[opened closed],
  merge_requests: %w[opened closed merged]
}.with_indifferent_access.freeze
MILESTONE_TIMEBOX_VALUES =
%w[none any upcoming started].freeze
ITERATION_SELECTION_VALUES =
%w[none any].freeze
EpicsTriagingForProjectImpossibleError =
Class.new(StandardError)
MultiPolicyInInjectionModeError =
Class.new(StandardError)

Instance Attribute Summary collapse

Instance Method Summary collapse

Constructor Details

#initialize(policies:, options:, network_adapter_class: DEFAULT_NETWORK_ADAPTER, graphql_network_adapter_class: DEFAULT_GRAPHQL_ADAPTER) ⇒ Engine

Returns a new instance of Engine.



72
73
74
75
76
77
78
79
80
81
82
83
84
85
86
87
88
# File 'lib/gitlab/triage/engine.rb', line 72

def initialize(policies:, options:, network_adapter_class: DEFAULT_NETWORK_ADAPTER, graphql_network_adapter_class: DEFAULT_GRAPHQL_ADAPTER)
  options.host_url = policies.delete(:host_url) { options.host_url }
  options.api_version = policies.delete(:api_version) { 'v4' }
  options.dry_run = ENV['TEST'] == 'true' if options.dry_run.nil?

  @per_page = policies.delete(:per_page) { 100 }
  @policies = policies
  @options = options
  @network_adapter_class = network_adapter_class
  @graphql_network_adapter_class = graphql_network_adapter_class

  assert_options!

  @options.source = @options.source.to_s

  require_ruby_files
end

Instance Attribute Details

#optionsObject (readonly)

Returns the value of attribute options.



39
40
41
# File 'lib/gitlab/triage/engine.rb', line 39

def options
  @options
end

#per_pageObject (readonly)

Returns the value of attribute per_page.



39
40
41
# File 'lib/gitlab/triage/engine.rb', line 39

def per_page
  @per_page
end

#policiesObject (readonly)

Returns the value of attribute policies.



39
40
41
# File 'lib/gitlab/triage/engine.rb', line 39

def policies
  @policies
end

Instance Method Details

#apply_work_item_status!(resource_type, expanded_conditions, rule_definition, resources, status_pre_filtered) ⇒ Object (private)



685
686
687
688
689
690
691
692
693
694
695
# File 'lib/gitlab/triage/engine.rb', line 685

def apply_work_item_status!(resource_type, expanded_conditions, rule_definition, resources, status_pre_filtered)
  return unless resource_type == 'issues' && needs_work_item_status?(expanded_conditions, rule_definition)

  status_values = Array(expanded_conditions[:work_item_status])

  if status_pre_filtered && status_values.size == 1
    resources.each { |r| r[:work_item_status] = status_values.first }
  else
    decorate_resources_with_work_item_status(resources)
  end
end

#assert_all!Object (private)

rubocop:disable Style/IfUnlessModifier



131
132
133
134
135
136
137
138
139
140
141
142
143
144
145
# File 'lib/gitlab/triage/engine.rb', line 131

def assert_all!
  return unless options.all

  if options.source
    raise ArgumentError, '--all-projects option cannot be used in conjunction with --source option!'
  end

  if options.source_id
    raise ArgumentError, '--all-projects option cannot be used in conjunction with --source-id option!'
  end

  if options.resource_reference # rubocop:disable Style/GuardClause
    raise ArgumentError, '--all-projects option cannot be used in conjunction with --resource-reference option!'
  end
end

#assert_epic_rule!(resource_type) ⇒ Object (private)



197
198
199
200
201
# File 'lib/gitlab/triage/engine.rb', line 197

def assert_epic_rule!(resource_type)
  return if resource_type != 'epics' || options.source == 'groups'

  raise EpicsTriagingForProjectImpossibleError, "Epics can only be triaged at the group level. Please set the `--source groups` option."
end

#assert_options!Object (private)



123
124
125
126
127
128
# File 'lib/gitlab/triage/engine.rb', line 123

def assert_options!
  assert_all!
  assert_source!
  assert_source_id!
  assert_resource_reference!
end

#assert_resource_reference!Object (private)



162
163
164
165
166
167
168
169
170
171
172
173
174
175
176
# File 'lib/gitlab/triage/engine.rb', line 162

def assert_resource_reference!
  return unless options.resource_reference

  if options.source == 'groups' && !options.resource_reference.start_with?('&')
    raise ArgumentError, "--resource-reference can only start with '&' when --source=groups is passed ('#{options.resource_reference}' passed)!"
  end

  if options.source == 'projects' && !options.resource_reference.start_with?('#', '!') # rubocop:disable Style/GuardClause
    raise(
      ArgumentError,
      "--resource-reference can only start with '#' or '!' when --source=projects is passed " \
        "('#{options.resource_reference}' passed)!"
    )
  end
end

#assert_source!Object (private)

rubocop:enable Style/IfUnlessModifier

Raises:

  • (ArgumentError)


148
149
150
151
152
153
# File 'lib/gitlab/triage/engine.rb', line 148

def assert_source!
  return if options.source
  return if options.all

  raise ArgumentError, 'A source is needed (pass it with the `--source` option)!'
end

#assert_source_id!Object (private)

Raises:

  • (ArgumentError)


155
156
157
158
159
160
# File 'lib/gitlab/triage/engine.rb', line 155

def assert_source_id!
  return if options.source_id
  return if options.all

  raise ArgumentError, 'A project or group ID is needed (pass it with the `--source-id` option)!'
end

#attach_resource_type(resources, resource_type) ⇒ Object (private)



449
450
451
# File 'lib/gitlab/triage/engine.rb', line 449

def attach_resource_type(resources, resource_type)
  resources.each { |resource| resource[:type] = resource_type }
end

#branches_resource_query(conditions) ⇒ Object (private)



646
647
648
649
650
# File 'lib/gitlab/triage/engine.rb', line 646

def branches_resource_query(conditions)
  [].tap do |condition_builders|
    condition_builders << APIQueryBuilders::SingleQueryParamBuilder.new('search', conditions[:name]) if conditions[:name]
  end
end

#build_get_url(resource_type, conditions) ⇒ Object (private)

rubocop:disable Metrics/AbcSize rubocop:disable Metrics/CyclomaticComplexity rubocop:disable Metrics/PerceivedComplexity



534
535
536
537
538
539
540
541
542
543
544
545
546
547
548
549
550
551
552
553
554
555
556
557
558
559
560
561
562
563
564
565
566
567
568
569
570
571
572
573
574
575
576
577
578
579
580
581
582
583
584
585
586
587
588
589
590
591
592
# File 'lib/gitlab/triage/engine.rb', line 534

def build_get_url(resource_type, conditions)
  # Example issues query with state and labels
  # https://gitlab.com/api/v4/projects/test-triage%2Fissue-project/issues?state=open&labels=project%20label%20with%20spaces,group_label_no_spaces
  params = {
    per_page: per_page
  }

  condition_builders = []
  condition_builders << APIQueryBuilders::SingleQueryParamBuilder.new('iids', options.resource_reference[1..]) if options.resource_reference
  author_username = conditions[:author_username]
  condition_builders << APIQueryBuilders::SingleQueryParamBuilder.new('author_username', author_username) if author_username

  condition_builders << APIQueryBuilders::MultiQueryParamBuilder.new('labels', conditions[:labels], ',') if conditions[:labels]

  if conditions[:forbidden_labels]
    condition_builders << APIQueryBuilders::MultiQueryParamBuilder.new('not[labels]', conditions[:forbidden_labels], ',')
  end

  if conditions[:state]
    condition_builders << APIQueryBuilders::SingleQueryParamBuilder.new(
      'state',
      conditions[:state],
      allowed_values: ALLOWED_STATE_VALUES[resource_type])
  end

  condition_builders << milestone_condition_builder(resource_type, conditions[:milestone]) if conditions[:milestone]

  if conditions[:date] && APIQueryBuilders::DateQueryParamBuilder.applicable?(conditions[:date]) && resource_type&.to_sym != :branches
    condition_builders << APIQueryBuilders::DateQueryParamBuilder.new(conditions.delete(:date))
  end

  case resource_type&.to_sym
  when :issues
    condition_builders.concat(issues_resource_query(conditions))
  when :merge_requests
    condition_builders.concat(merge_requests_resource_query(conditions))
  when :branches
    condition_builders.concat(branches_resource_query(conditions))
  end

  condition_builders.compact.each do |condition_builder|
    params[condition_builder.param_name] = condition_builder.param_content
  end

  url_builder_options = {
    network_options: options,
    all: options.all,
    source: options.source,
    source_id: options.source_id,
    resource_type: resource_type,
    params: params
  }

  # FIXME: Epics listing endpoint doesn't support filtering by `iids`, so instead we
  # get a single epic when `--resource-reference` is given for epics.
  url_builder_options[:resource_id] = options.resource_reference[1..] if options.resource_reference && resource_type == 'epics'

  UrlBuilders::UrlBuilder.new(url_builder_options).build
end

#build_graphql_query(resource_type, conditions, graphql_only = false) ⇒ Object (private)



822
823
824
825
# File 'lib/gitlab/triage/engine.rb', line 822

def build_graphql_query(resource_type, conditions, graphql_only = false)
  Gitlab::Triage::GraphqlQueries::QueryBuilder
    .new(options.source, resource_type, conditions, graphql_only: graphql_only)
end

#decorate_resources_with_graphql_data(resources, graphql_resources) ⇒ Object (private)



453
454
455
456
457
458
# File 'lib/gitlab/triage/engine.rb', line 453

def decorate_resources_with_graphql_data(resources, graphql_resources)
  return if graphql_resources.nil?

  graphql_resources_by_id = graphql_resources.index_by { |resource| resource[:id] }
  resources.each { |resource| resource.merge!(graphql_resources_by_id[resource[:id]].to_h) }
end

#decorate_resources_with_work_item_status(resources) ⇒ Object (private)

Status is a work item capability not available on the IssueType GraphQL type or the REST API. We fetch it via a separate workItems query and merge it back by IID, similar to how decorate_resources_with_graphql_data works for other fields. We use widgets (not features) for backward compatibility with GitLab versions before the features field was introduced in 18.9.



754
755
756
757
758
759
760
761
762
763
764
765
766
# File 'lib/gitlab/triage/engine.rb', line 754

def decorate_resources_with_work_item_status(resources)
  return if resources.empty?

  iids = resources.filter_map { |r| (r['iid'] || r[:iid])&.to_s }
  return if iids.empty?

  status_by_iid = fetch_work_item_statuses_by_iid(iids)

  resources.each do |resource|
    iid = (resource['iid'] || resource[:iid]).to_s
    resource[:work_item_status] = status_by_iid[iid]
  end
end

#draft_condition_builder(draft_condittion) ⇒ Object (private)



652
653
654
655
656
657
658
659
660
661
662
663
664
665
# File 'lib/gitlab/triage/engine.rb', line 652

def draft_condition_builder(draft_condittion)
  # Issues API only accepts 'yes' and 'no' as strings: https://docs.gitlab.com/ee/api/merge_requests.html
  wip =
    case draft_condittion
    when true
      'yes'
    when false
      'no'
    else
      raise ArgumentError, 'The "draft" condition only accepts true or false.'
    end

  APIQueryBuilders::SingleQueryParamBuilder.new('wip', wip)
end

#fetch_resources(resource_type, expanded_conditions, rule_definition) ⇒ Object (private)



409
410
411
412
413
414
415
416
417
418
419
420
421
422
423
424
425
426
427
428
429
430
431
432
433
434
435
436
437
438
439
440
441
442
443
444
445
446
447
# File 'lib/gitlab/triage/engine.rb', line 409

def fetch_resources(resource_type, expanded_conditions, rule_definition)
  resources = []

  if rule_definition[:api] == 'graphql'
    graphql_query_options = { source: source_full_path }

    if options.resource_reference
      expanded_conditions[:iids] = options.resource_reference[1..]
      graphql_query_options[:iids] = [expanded_conditions[:iids]]
    elsif expanded_conditions[:iids].present?
      graphql_query_options[:iids] = expanded_conditions[:iids]
    end

    graphql_query = build_graphql_query(resource_type, expanded_conditions, true)

    resources = graphql_network.query(graphql_query, **graphql_query_options)
  else
    # FIXME: Epics listing endpoint doesn't support filtering by `iids`, so instead we
    # get a single epic when `--resource-reference` is given for epics.
    # Because of that, the query could return a single epic, so we make sure we get an array.
    pre_filter_iids = expanded_conditions.delete(:iids)

    resources = Array(network.query_api(build_get_url(resource_type, expanded_conditions)))

    # When pre-filtered IIDs are present (from work_item_status server-side
    # filtering), narrow the REST results to only those IIDs before decoration.
    resources = resources.select { |r| pre_filter_iids.include?(r['iid'].to_s) } if pre_filter_iids.present?

    iids = resources.pluck('iid').map(&:to_s)
    expanded_conditions[:iids] = iids

    graphql_query = build_graphql_query(resource_type, expanded_conditions)
    graphql_resources = graphql_network.query(graphql_query, source: source_full_path, iids: iids) if graphql_query.any?

    decorate_resources_with_graphql_data(resources, graphql_resources)
  end

  resources
end

#fetch_source_full_pathObject (private)

Raises:

  • (ArgumentError)


831
832
833
834
835
836
837
838
839
840
# File 'lib/gitlab/triage/engine.rb', line 831

def fetch_source_full_path
  return options.source_id unless /\A\d+\z/.match?(options.source_id)

  source_details = network.query_api(build_get_url(nil, {})).first
  full_path = source_details['full_path'] || source_details['path_with_namespace']

  raise ArgumentError, 'A source with given source_id was not found!' if full_path.blank?

  full_path
end

#fetch_work_item_iids_by_status(status_condition) ⇒ Object (private)

Pre-filter: query workItems with server-side status filtering to get only the IIDs that match, avoiding fetching the full resource set when most items won’t pass the status filter.



704
705
706
707
708
709
710
711
712
713
714
715
716
717
718
719
720
721
722
723
724
725
726
727
728
729
730
731
732
733
734
735
736
737
738
739
740
741
742
743
744
745
746
747
# File 'lib/gitlab/triage/engine.rb', line 704

def fetch_work_item_iids_by_status(status_condition)
  status_names = Array(status_condition)
  all_iids = []

  status_names.each do |status_name|
    query = <<~GRAPHQL
      query($source: ID!, $after: String, $statusName: String!) {
        #{options.source.singularize}(fullPath: $source) {
          workItems(status: { name: $statusName }, after: $after, first: 100) {
            pageInfo {
              hasNextPage
              endCursor
            }
            nodes {
              iid
            }
          }
        }
      }
    GRAPHQL

    after_cursor = nil

    loop do
      response = graphql_network.adapter.query_raw(
        query,
        resource_path: [options.source.singularize, 'workItems'],
        variables: { source: source_full_path, after: after_cursor, statusName: status_name }
      )

      work_items = Array.wrap(response[:results])
      work_items.each do |wi|
        iid = wi.is_a?(Hash) ? (wi['iid'] || wi[:iid]) : wi.try(:iid)
        all_iids << iid.to_s if iid
      end

      break unless response[:more_pages]

      after_cursor = response[:end_cursor]
    end
  end

  all_iids.uniq
end

#fetch_work_item_statuses_by_iid(iids) ⇒ Object (private)

The workItems connection caps ‘first` at 100, so we request the status for the given IIDs in batches and paginate within each batch.



770
771
772
773
774
775
776
777
778
779
780
781
782
783
784
785
786
787
788
789
790
791
792
793
794
795
796
797
798
799
800
801
802
803
804
805
806
807
808
809
810
811
812
813
814
815
816
817
818
819
820
# File 'lib/gitlab/triage/engine.rb', line 770

def fetch_work_item_statuses_by_iid(iids)
  query = <<~GRAPHQL
    query($source: ID!, $iids: [String!], $after: String) {
      #{options.source.singularize}(fullPath: $source) {
        workItems(iids: $iids, after: $after, first: 100) {
          pageInfo {
            hasNextPage
            endCursor
          }
          nodes {
            iid
            widgets(onlyTypes: [STATUS]) {
              ... on WorkItemWidgetStatus {
                status {
                  id
                  name
                }
              }
            }
          }
        }
      }
    }
  GRAPHQL

  status_by_iid = {}

  iids.each_slice(100) do |iids_batch|
    after_cursor = nil

    loop do
      response = graphql_network.adapter.query_raw(
        query,
        resource_path: [options.source.singularize, 'workItems'],
        variables: { source: source_full_path, iids: iids_batch, after: after_cursor }
      )

      Array.wrap(response[:results]).each do |wi|
        wi = wi.deep_transform_keys(&:underscore).with_indifferent_access
        status_widget = Array.wrap(wi[:widgets]).find { |w| w[:status].present? }
        status_by_iid[wi[:iid].to_s] = status_widget.dig(:status, :name) if status_widget
      end

      break unless response[:more_pages]

      after_cursor = response[:end_cursor]
    end
  end

  status_by_iid
end

#filter_resource(resource, conditions) ⇒ Object (private)



474
475
476
477
478
479
480
481
482
483
484
485
486
487
488
489
490
491
492
493
494
495
496
497
498
499
500
501
502
503
504
505
506
507
508
509
510
511
512
513
514
# File 'lib/gitlab/triage/engine.rb', line 474

def filter_resource(resource, conditions)
  results = []

  FILTER_MAP.each do |condition_key, filter_value|
    # Skips to the next key value pair if the condition is not applicable
    next if conditions[condition_key].nil?

    case filter_value
    when Hash
      filter_in_ruby = conditions[condition_key].dig(:filter_in_ruby)
      merged_at = conditions[condition_key].dig(:attribute) == 'merged_at'
      filter_branch = conditions.dig(:date) && resource[:type] == 'branches'

      # Set the filter to the resource type
      if filter_in_ruby || merged_at || filter_branch
        filter = filter_value[resource[:type]]
        results << filter.new(resource, conditions[condition_key]).calculate
      end
    else
      # The `filter_value` set is not of type `hash`
      filter = filter_value

      # If the :ruby condition exists then filter based off of conditions
      # else we base off of the `conditions[condition_key]`.

      result =
        if condition_key.to_s == 'no_additional_labels'
          filter.new(resource, conditions[:labels]).calculate
        elsif condition_key.to_s == 'protected'
          filter.new(resource, conditions[:protected]).calculate
        elsif filter.instance_method(:initialize).arity == 2
          filter.new(resource, conditions[condition_key]).calculate
        else
          filter.new(resource, conditions[condition_key], network).calculate
        end

      results << result
    end
  end
  results.all?
end

#filter_resources(resources, conditions) ⇒ Object (private)



468
469
470
471
472
# File 'lib/gitlab/triage/engine.rb', line 468

def filter_resources(resources, conditions)
  resources.select do |resource|
    filter_resource(resource, conditions)
  end
end

#graphql_networkObject



117
118
119
# File 'lib/gitlab/triage/engine.rb', line 117

def graphql_network
  @graphql_network ||= GraphqlNetwork.new(graphql_network_adapter)
end

#graphql_network_adapterObject (private)



211
212
213
# File 'lib/gitlab/triage/engine.rb', line 211

def graphql_network_adapter
  @graphql_network_adapter ||= @graphql_network_adapter_class.new(options)
end

#issues_resource_query(conditions) ⇒ Object (private)



624
625
626
627
628
629
630
631
# File 'lib/gitlab/triage/engine.rb', line 624

def issues_resource_query(conditions)
  [].tap do |condition_builders|
    condition_builders << APIQueryBuilders::SingleQueryParamBuilder.new('weight', conditions[:weight]) if conditions[:weight]
    condition_builders << iteration_condition_builder(conditions[:iteration]) if conditions[:iteration]
    condition_builders << APIQueryBuilders::SingleQueryParamBuilder.new('health_status', conditions[:health_status]) if conditions[:health_status]
    condition_builders << APIQueryBuilders::SingleQueryParamBuilder.new('issue_type', conditions[:issue_type]) if conditions[:issue_type]
  end
end

#iteration_condition_builder(iteration_value) ⇒ Object (private)



612
613
614
615
616
617
618
619
620
621
622
# File 'lib/gitlab/triage/engine.rb', line 612

def iteration_condition_builder(iteration_value)
  # Issues API should use the `iteration_id` param for timebox values, and `iteration_title` for iteration title
  args =
    if ITERATION_SELECTION_VALUES.include?(iteration_value.downcase)
      ['iteration_id', iteration_value.titleize] # The API only accepts titleized values.
    else
      ['iteration_title', iteration_value]
    end

  APIQueryBuilders::SingleQueryParamBuilder.new(*args)
end

#limit_resources(resources, limits) ⇒ Object (private)



516
517
518
519
520
521
522
# File 'lib/gitlab/triage/engine.rb', line 516

def limit_resources(resources, limits)
  if limits.empty?
    resources
  else
    Limiters::DateFieldLimiter.new(resources, limits).limit
  end
end

#merge_requests_resource_query(conditions) ⇒ Object (private)



633
634
635
636
637
638
639
640
641
642
643
644
# File 'lib/gitlab/triage/engine.rb', line 633

def merge_requests_resource_query(conditions)
  [].tap do |condition_builders|
    [
      :source_branch,
      :target_branch,
      :reviewer_id
    ].each do |key|
      condition_builders << APIQueryBuilders::SingleQueryParamBuilder.new(key.to_s, conditions[key]) if conditions[key]
    end
    condition_builders << draft_condition_builder(conditions[:draft]) if conditions.key?(:draft)
  end
end

#milestone_condition_builder(resource_type, milestone_condition) ⇒ Object (private)

rubocop:enable Metrics/AbcSize rubocop:enable Metrics/CyclomaticComplexity rubocop:enable Metrics/PerceivedComplexity



597
598
599
600
601
602
603
604
605
606
607
608
609
610
# File 'lib/gitlab/triage/engine.rb', line 597

def milestone_condition_builder(resource_type, milestone_condition)
  milestone_value = Array(milestone_condition)[0].to_s # back-compatibility
  return if milestone_value.empty?

  # Issues API should use the `milestone_id` param for timebox values, and `milestone` for milestone title
  args =
    if resource_type.to_sym == :issues && MILESTONE_TIMEBOX_VALUES.include?(milestone_value.downcase)
      ['milestone_id', milestone_value.titleize] # The API only accepts titleized values.
    else
      ['milestone', milestone_value]
    end

  APIQueryBuilders::SingleQueryParamBuilder.new(*args)
end

#needs_work_item_status?(conditions, _rule_definition) ⇒ Boolean (private)

Returns:

  • (Boolean)


697
698
699
# File 'lib/gitlab/triage/engine.rb', line 697

def needs_work_item_status?(conditions, _rule_definition)
  conditions.key?(:work_item_status)
end

#networkObject



109
110
111
# File 'lib/gitlab/triage/engine.rb', line 109

def network
  @network ||= Network.new(restapi: restapi_network, graphql: graphql_network)
end

#network_adapterObject (private)



207
208
209
# File 'lib/gitlab/triage/engine.rb', line 207

def network_adapter
  @network_adapter ||= @network_adapter_class.new(options)
end

#performObject



90
91
92
93
94
95
96
97
98
99
100
101
102
103
104
105
106
107
# File 'lib/gitlab/triage/engine.rb', line 90

def perform
  puts "Performing a dry run.\n\n" if options.dry_run

  puts Gitlab::Triage::UI.header("Triaging the `#{options.source_id}` #{options.source.singularize}", char: '=')
  puts

  resource_rules.each do |resource_type, policy_definition|
    next unless right_resource_type_for_resource_option?(resource_type)

    assert_epic_rule!(resource_type)

    puts Gitlab::Triage::UI.header("Processing summaries & rules for #{resource_type}", char: '-')
    puts

    process_summaries(resource_type, policy_definition[:summaries])
    process_rules(resource_type, policy_definition[:rules])
  end
end

#pre_filter_by_work_item_status!(resource_type, expanded_conditions) ⇒ Object (private)

When filtering by work_item_status, query workItems with server-side status filtering first to narrow the IID set before the main fetch. This avoids fetching all resources just to discard most of them.



670
671
672
673
674
675
676
677
678
679
680
681
682
683
# File 'lib/gitlab/triage/engine.rb', line 670

def pre_filter_by_work_item_status!(resource_type, expanded_conditions)
  return false unless resource_type == 'issues' && expanded_conditions.key?(:work_item_status)

  pre_filter_iids = fetch_work_item_iids_by_status(expanded_conditions[:work_item_status])

  expanded_conditions[:iids] =
    if expanded_conditions[:iids].present?
      Array(expanded_conditions[:iids]).map(&:to_s) & pre_filter_iids
    else
      pre_filter_iids
    end

  true
end

#process_action(policy) ⇒ Object (private)



460
461
462
463
464
465
466
# File 'lib/gitlab/triage/engine.rb', line 460

def process_action(policy)
  Action.process(
    policy: policy,
    network: network,
    dry: options.dry_run)
  puts
end

#process_rules(resource_type, rule_definitions) ⇒ nil (private)

Process an array of rule_definitions.

Examples:

Example of an array of rule definitions.


[{ name: "New issues", conditions: { state: opened }, limits: { most_recent: 2 }, actions: { labels: ["needs attention"] } }]

Parameters:

  • rule_definitions (Array<Hash>)

    An array usually given as YAML in a triage policy file.

Returns:

  • (nil)


269
270
271
272
273
274
275
276
277
278
279
280
# File 'lib/gitlab/triage/engine.rb', line 269

def process_rules(resource_type, rule_definitions)
  return if rule_definitions.blank?

  rule_definitions.each do |rule_definition|
    resources_for_rule(resource_type, rule_definition, in_summary: false) do |resources|
      policy = Policies::RulePolicy.new(
        resource_type, rule_definition, resources, network)

      process_action(policy)
    end
  end
end

#process_summaries(resource_type, summary_definitions) ⇒ nil (private)

Process an array of summary_definitions.

Examples:

Example of an array of summary definitions (shown as YAML for readability).


- name: Newest and oldest issues summary
  rules:
    - name: New issues
      conditions:
        state: opened
      limits:
        most_recent: 2
      actions:
        summarize:
          item: "- [ ] [{{title}}]({{web_url}}) {{labels}}"
          summary: |
            Please triage the following new {{type}}:
            {{items}}
  actions:
    summarize:
      title: "Newest and oldest {{type}} summary"
      summary: |
        Please triage the following {{type}}:
        {{items}}
        Please take care of them before the end of #{7.days.from_now.strftime('%Y-%m-%d')}
        /label ~"needs attention"

Parameters:

  • summary_definitions (Array<Hash>)

    An array usually given as YAML in a triage policy file.

Returns:

  • (nil)


252
253
254
255
256
257
258
# File 'lib/gitlab/triage/engine.rb', line 252

def process_summaries(resource_type, summary_definitions)
  return if summary_definitions.blank?

  summary_definitions.each do |summary_definition|
    process_summary(resource_type, summary_definition)
  end
end

#process_summary(resource_type, summary_definition) ⇒ nil (private)

Process a summary_definition.

Examples:

Example of a summary definition hash (shown as YAML for readability).


name: Newest and oldest issues summary
rules:
  - name: New issues
    conditions:
      state: opened
    limits:
      most_recent: 2
    actions:
      summarize:
        item: "- [ ] [{{title}}]({{web_url}}) {{labels}}"
        summary: |
          Please triage the following new {{type}}:
          {{items}}
actions:
  summarize:
    title: "Newest and oldest {{type}} summary"
    summary: |
      Please triage the following {{type}}:
      {{items}}
      Please take care of them before the end of #{7.days.from_now.strftime('%Y-%m-%d')}
      /label ~"needs attention"

Parameters:

  • resource_type (String)

    The resource type, e.g. issues or merge_requests.

  • summary_definition (Hash)

    A hash usually given as YAML in a triage policy file:

Returns:

  • (nil)


312
313
314
315
316
317
318
319
320
321
322
# File 'lib/gitlab/triage/engine.rb', line 312

def process_summary(resource_type, summary_definition)
  puts Gitlab::Triage::UI.header("Processing summary: **#{summary_definition[:name]}**", char: '~')
  puts

  summary_parts_for_rules(resource_type, summary_definition[:rules]) do |summary_resources|
    policy = Policies::SummaryPolicy.new(
      resource_type, summary_definition, summary_resources, network)

    process_action(policy)
  end
end

#require_ruby_filesObject (private)



178
179
180
# File 'lib/gitlab/triage/engine.rb', line 178

def require_ruby_files
  options.require_files.each(&method(:require))
end

#resource_rulesObject (private)



203
204
205
# File 'lib/gitlab/triage/engine.rb', line 203

def resource_rules
  @resource_rules ||= policies.delete(:resource_rules) { {} }
end

#resources_for_rule(resource_type, rule_definition, in_summary: false) {|rule_resources, expanded_conditions| ... } ⇒ nil (private)

Transform a non-expanded rule_definition into a PoliciesResources::RuleResources.new(resources) object.

Examples:

Example of a rule definition hash.


{ name: "New issues", conditions: { state: opened }, limits: { most_recent: 2 }, actions: { labels: ["needs attention"] } }

Parameters:

  • resource_type (String)

    The resource type, e.g. issues or merge_requests.

  • rule_definition (Hash)

    A rule definition, e.g. { name: ‘Foo’, conditions: { milestone: ‘v1’ } }.

  • in_summary (Boolean) (defaults to: false)

    Whether this rule is part of a summary definition.

Yield Parameters:

  • rule_resources (PoliciesResources::RuleResources)

    An object which contains an array of resources.

  • expanded_conditions (Hash)

    A hash of expanded conditions.

Yield Returns:

  • (nil)

Returns:

  • (nil)


367
368
369
370
371
372
373
374
375
376
377
378
379
380
381
382
383
384
385
386
387
388
389
390
391
392
393
394
395
396
397
398
399
400
401
402
403
404
405
406
407
# File 'lib/gitlab/triage/engine.rb', line 367

def resources_for_rule(resource_type, rule_definition, in_summary: false)
  puts Gitlab::Triage::UI.header("Gathering resources for rule: **#{rule_definition[:name]}**", char: '-')

  # Skip resource fetching for custom type rules
  if rule_definition[:type] == 'custom'
    unless in_summary
      raise ArgumentError, "type: custom can only be used in summary rules, not in top-level rules (rule: #{rule_definition[:name]})"
    end

    puts "\n* Skipping resource fetch for custom type rule"
    yield(PoliciesResources::RuleResources.new([]), {})
    return
  end

  ExpandCondition.perform(rule_conditions(rule_definition)) do |expanded_conditions|
    status_pre_filtered = pre_filter_by_work_item_status!(resource_type, expanded_conditions)

    # retrieving the resources for every rule is inefficient
    # however, previous rules may affect those upcoming
    resources = options.resources ||
      fetch_resources(resource_type, expanded_conditions, rule_definition)

    apply_work_item_status!(resource_type, expanded_conditions, rule_definition, resources, status_pre_filtered)

    # In some filters/actions we want to know which resource type it is
    attach_resource_type(resources, resource_type)

    puts "\n\n* Found #{resources.count} resources..."
    print "* Filtering resources..."
    resources = filter_resources(resources, expanded_conditions)
    puts "\n* Total after filtering: #{resources.count} resources"
    print "* Limiting resources..."
    resources = limit_resources(resources, rule_limits(rule_definition))
    puts "\n* Total after limiting: #{resources.count} resources"
    puts

    resources = sanitize_resources(resources)

    yield(PoliciesResources::RuleResources.new(resources), expanded_conditions)
  end
end

#restapi_networkObject



113
114
115
# File 'lib/gitlab/triage/engine.rb', line 113

def restapi_network
  @restapi_network ||= RestAPINetwork.new(network_adapter)
end

#right_resource_type_for_resource_option?(resource_type) ⇒ Boolean (private)

Returns:

  • (Boolean)


182
183
184
185
186
187
188
189
190
191
192
193
194
195
# File 'lib/gitlab/triage/engine.rb', line 182

def right_resource_type_for_resource_option?(resource_type)
  return true unless options.resource_reference

  resource_reference = options.resource_reference

  case resource_type
  when 'issues'
    resource_reference.start_with?('#')
  when 'merge_requests'
    resource_reference.start_with?('!')
  when 'epics'
    resource_reference.start_with?('&')
  end
end

#rule_conditions(rule) ⇒ Object (private)



215
216
217
# File 'lib/gitlab/triage/engine.rb', line 215

def rule_conditions(rule)
  rule.fetch(:conditions) { {} }
end

#rule_limits(rule) ⇒ Object (private)



219
220
221
# File 'lib/gitlab/triage/engine.rb', line 219

def rule_limits(rule)
  rule.fetch(:limits) { {} }
end

#sanitize_resources(resources) ⇒ Object (private)



524
525
526
527
528
529
# File 'lib/gitlab/triage/engine.rb', line 524

def sanitize_resources(resources)
  resources.each do |resource|
    # Titles should not contain newlines. Translate them to spaces.
    resource[:title] = resource[:title]&.tr("\r\n", '  ')
  end
end

#source_full_pathObject (private)



827
828
829
# File 'lib/gitlab/triage/engine.rb', line 827

def source_full_path
  @source_full_path ||= fetch_source_full_path
end

#summary_parts_for_rules(resource_type, rule_definitions) {|summary_resources| ... } ⇒ nil (private)

Transform an array of rule_definitions into a PoliciesResources::SummaryResources.new(rule => rule_resources) object.

Examples:

Example of an array of rule definitions.


[{ name: "New issues", conditions: { state: opened }, limits: { most_recent: 2 }, actions: { labels: ["needs attention"] } }]

Parameters:

  • resource_type (String)

    The resource type, e.g. issues or merge_requests.

  • rule_definitions (Array<Hash>)

    An array of rule definitions, e.g. [{ name: ‘Foo’, conditions: { milestone: ‘v1’ } }, { name: ‘Foo’, conditions: { state: ‘opened’ } }].

Yield Parameters:

Yield Returns:

  • (nil)

Returns:

  • (nil)


338
339
340
341
342
343
344
345
346
347
348
349
350
# File 'lib/gitlab/triage/engine.rb', line 338

def summary_parts_for_rules(resource_type, rule_definitions)
  # { summary_rule => resources }
  parts = rule_definitions.each_with_object({}) do |rule_definition, result|
    to_enum(:resources_for_rule, resource_type, rule_definition, in_summary: true).each do |rule_resources, expanded_conditions|
      # We replace the non-expanded rule conditions with the expanded ones
      result.merge!(rule_definition.merge(conditions: expanded_conditions) => rule_resources)
    end

    result
  end

  yield(PoliciesResources::SummaryResources.new(parts))
end